"""Utility functions for Dapla Pseudo."""

import asyncio
import typing as t
from datetime import date
from pathlib import Path

from pydantic import ValidationError

from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.exceptions import NoFileExtensionError
from dapla_pseudo.v1.models.api import DepseudoFieldRequest
from dapla_pseudo.v1.models.api import PseudoFieldRequest
from dapla_pseudo.v1.models.api import RawPseudoMetadata
from dapla_pseudo.v1.models.api import RepseudoFieldRequest
from dapla_pseudo.v1.models.core import KeyWrapper
from dapla_pseudo.v1.models.core import PseudoKeyset
from dapla_pseudo.v1.models.core import PseudoRule
from dapla_pseudo.v1.models.core import RedactKeywordArgs
from dapla_pseudo.v1.mutable_dataframe import MutableDataFrame
from dapla_pseudo.v1.supported_file_format import SupportedOutputFileFormat


def find_multipart_obj(obj_name: str, multipart_files_tuple: set[t.Any]) -> t.Any:
    """Find "multipart object" by name.

    The requests lib specifies multipart file arguments as file-tuples, such as
    ('filename', fileobj, 'content_type')
    This method searches a tuple of such file-tuples ((file-tuple1),...,(file-tupleN))
    It returns the fileobj for the first matching file-tuple with a specified filename.

    Example:
    Given the multipart_files_tuple:
    multipart_tuple = (('filename1', fileobj1, 'application/json'), ('filename2', fileobj2, 'application/json'))

    then
    find_multipart_obj("filename2", multipart_tuple) -> fileobj
    """
    try:
        matching_item = next(
            item[1] for item in multipart_files_tuple if item[0] == obj_name
        )
        return matching_item[1]
    except StopIteration:
        return None


def redact_field(
    request: PseudoFieldRequest,
) -> tuple[str, list[str | None], RawPseudoMetadata]:
    """Perform the redact operation locally.

    This is in order to avoid making unnecessary requests to the API.
    """
    kwargs = t.cast(RedactKeywordArgs, request.pseudo_func.kwargs)
    if kwargs.placeholder is None:
        raise ValueError("Placeholder needs to be set for Redact")
    data: list[str | None] = [kwargs.placeholder for _ in request.values]
    # The above operation could be vectorized using something like Polars,
    # however - the redact functionality is used mostly teams that use hierarchical
    # data, i.e. with very small lists. The overhead of
    # creating a Polars Series is probably not worth it.

    metadata = RawPseudoMetadata(
        field_name=request.name,
        logs=[],
        metrics=[],
        datadoc=[
            {
                "short_name": request.name.split("/")[-1],
                "data_element_path": request.name.replace("/", "."),
                "pseudonymization": {
                    "encryption_algorithm": "REDACT",
                    "encryption_algorithm_parameters": [
                        request.pseudo_func.kwargs.model_dump(exclude_none=True)
                    ],
                },
            }
        ],
    )

    return request.name, data, metadata


def running_asyncio_loop() -> asyncio.AbstractEventLoop | None:
    """Returns the asyncio event loop if it exists."""
    try:
        loop = asyncio.get_running_loop()
        if loop.is_running():
            return loop
        else:
            return None
    except RuntimeError:
        return None


def convert_to_date(sid_snapshot_date: date | str | None = None) -> date | None:
    """Converts the SID version date to the 'date' type, if it is a string.

    If None, simply passes the None through the function.
    """
    if isinstance(sid_snapshot_date, str):
        try:
            return date.fromisoformat(sid_snapshot_date)
        except ValueError as exc:
            raise ValueError(
                "Version timestamp must be a valid ISO date string (YYYY-MM-DD)"
            ) from exc
    return sid_snapshot_date


def get_file_format_from_file_name(file_path: str | Path) -> SupportedOutputFileFormat:
    """Extracts the file format from a file path."""
    if isinstance(file_path, str):
        file_path = Path(file_path)

    file_extension = file_path.suffix
    if not file_extension:
        raise NoFileExtensionError(f"File path '{file_path}' has no file extension.")
    file_format = SupportedOutputFileFormat(file_extension.replace(".", ""))
    return file_format


def build_pseudo_field_request(
    pseudo_operation: PseudoOperation,
    mutable_df: MutableDataFrame,
    rules: list[PseudoRule],  # "source rules" if repseudo
    custom_keyset: PseudoKeyset | str | None = None,
    target_custom_keyset: PseudoKeyset | str | None = None,  # used in repseudo
    target_rules: list[PseudoRule] | None = None,  # used in repseudo)
) -> list[PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest]:
    """Builds a FieldRequest object."""
    mutable_df.match_rules(rules, target_rules)
    matched_fields = mutable_df.get_matched_fields()
    requests: list[PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest] = (
        []
    )
    req: PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest
    match pseudo_operation:
        case PseudoOperation.PSEUDONYMIZE:
            for field in matched_fields.values():
                try:
                    req = PseudoFieldRequest(
                        pseudo_func=field.func,
                        name=field.path,
                        pattern=field.pattern,
                        values=field.get_value(),
                        keyset=KeyWrapper(custom_keyset).keyset,
                    )
                    requests.append(req)
                except ValidationError as e:
                    raise Exception(f"Path or column: {field.path}") from e
        case PseudoOperation.DEPSEUDONYMIZE:
            for field in matched_fields.values():
                try:
                    req = DepseudoFieldRequest(
                        pseudo_func=field.func,
                        name=field.path,
                        pattern=field.pattern,
                        values=field.get_value(),
                        keyset=KeyWrapper(custom_keyset).keyset,
                    )
                    requests.append(req)
                except ValidationError as e:
                    raise Exception(f"Path or column: {field.path}") from e

        case PseudoOperation.REPSEUDONYMIZE:
            if target_rules is not None:
                for field in matched_fields.values():
                    try:
                        req = RepseudoFieldRequest(
                            source_pseudo_func=field.func,
                            target_pseudo_func=field.target_func,
                            name=field.path,
                            pattern=field.pattern,
                            values=field.get_value(),
                            source_keyset=KeyWrapper(custom_keyset).keyset,
                            target_keyset=KeyWrapper(target_custom_keyset).keyset,
                        )
                        requests.append(req)
                    except ValidationError as e:
                        raise Exception(f"Path or column: {field.path}") from e
            else:
                raise ValueError("Found no target rules")
    return requests



"""This module defines constants that are referenced throughout the codebase."""

from enum import Enum

TIMEOUT_DEFAULT: int = 60 * 60 * 10  # seconds


class Env(str, Enum):
    """Environment variable keys."""

    PSEUDO_SERVICE_URL = "PSEUDO_SERVICE_URL"
    PSEUDO_SERVICE_AUTH_TOKEN = "PSEUDO_SERVICE_AUTH_TOKEN"  # S105
    PSEUDO_CLIENT_ROWS_PER_PARTITION = "PSEUDO_CLIENT_ROWS_PER_PARTITION"
    PSEUDO_CLIENT_MAX_TOTAL_PARTITIONS = "PSEUDO_CLIENT_MAX_TOTAL_PARTITIONS"

    def __str__(self) -> str:
        """Use value for string representation."""
        return str(self.value)


class PseudoOperation(str, Enum):
    """Pseudo operation."""

    PSEUDONYMIZE = "pseudonymize"
    DEPSEUDONYMIZE = "depseudonymize"
    REPSEUDONYMIZE = "repseudonymize"


class PredefinedKeys(str, Enum):
    """Names of 'global keys' that the Dapla Pseudo Service is familiar with."""

    SSB_COMMON_KEY_1 = "ssb-common-key-1"
    SSB_COMMON_KEY_2 = "ssb-common-key-2"
    PAPIS_COMMON_KEY_1 = "papis-common-key-1"

    def __str__(self) -> str:
        """Use value for string representation."""
        return str(self.value)


class PseudoFunctionTypes(str, Enum):
    """Names of well known pseudo functions."""

    DAEAD = "daead"
    MAP_SID = "map-sid-ff31"
    FF31 = "ff31"
    REDACT = "redact"

    def __str__(self) -> str:
        """Use value for string representation."""
        return str(self.value)


class UnknownCharacterStrategy(str, Enum):
    """UnknownCharacterStrategy defines how encryption/decryption should handle non-alphabet characters."""

    FAIL = "fail"
    SKIP = "skip"
    DELETE = "delete"
    REDACT = "redact"

    def __str__(self) -> str:
        """Use value for string representation."""
        return str(self.value)


class MapFailureStrategy(str, Enum):
    """UnknownCharacterStrategy defines how encryption/decryption should handle non-alphabet characters."""

    RETURN_NULL = "RETURN_NULL"
    RETURN_ORIGINAL = "RETURN_ORIGINAL"

    def __str__(self) -> str:
        """Use value for string representation."""
        return str(self.value)



"""The models module contains base classes used by other models."""

from humps import camelize
from pydantic import BaseModel
from pydantic import ConfigDict


class APIModel(BaseModel):
    """APIModel is a base class for models that are used for communicating with the Dapla Pseudo Service.

    It provides configuration for serializing/converting between camelCase (required by the API) and
    snake_case (used pythonically by this lib). It also provides some good defaults for converting a
    model to JSON.
    """

    model_config = ConfigDict(alias_generator=camelize, populate_by_name=True)

    def to_json(self) -> str:
        """Convert the model to JSON using camelCase aliases and only including assigned values."""
        return self.model_dump_json(
            exclude_unset=True, exclude_none=True, by_alias=True
        )



"""Builder for submitting a pseudonymization request."""

from datetime import date
from typing import ClassVar

import pandas as pd
import polars as pl
from dapla_metadata.datasets.core import Datadoc

from dapla_pseudo.constants import TIMEOUT_DEFAULT
from dapla_pseudo.constants import MapFailureStrategy
from dapla_pseudo.constants import PredefinedKeys
from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.v1.baseclasses import _BasePseudonymizer
from dapla_pseudo.v1.baseclasses import _BaseRuleConstructor
from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoKeyset
from dapla_pseudo.v1.models.core import PseudoRule
from dapla_pseudo.v1.result import Result


class Pseudonymize:
    """Starting point for pseudonymization of datasets.

    This class should not be instantiated, only the static methods should be used.
    """

    dataset: pl.DataFrame

    @staticmethod
    def from_pandas(dataframe: pd.DataFrame) -> "Pseudonymize._Pseudonymizer":
        """Initialize a pseudonymization request from a Pandas DataFrame.

        Args:
            dataframe: A Pandas DataFrame

        Returns:
            _Pseudonymizer: An instance of the _Pseudonymizer class.
        """
        Pseudonymize.dataset = pl.from_pandas(dataframe)
        return Pseudonymize._Pseudonymizer()

    @staticmethod
    def from_polars(dataframe: pl.DataFrame) -> "Pseudonymize._Pseudonymizer":
        """Initialize a pseudonymization request from a Polars DataFrame.

        Args:
            dataframe: A Polars DataFrame

        Returns:
            _Pseudonymizer: An instance of the _Pseudonymizer class.
        """
        Pseudonymize.dataset = dataframe
        return Pseudonymize._Pseudonymizer()

    class _Pseudonymizer(_BasePseudonymizer):
        """Select one or multiple fields to be pseudonymized."""

        rules: ClassVar[list[PseudoRule]] = []
        metadata: Datadoc | None = None

        def __init__(
            self, rules: list[PseudoRule] | None = None, metadata: Datadoc | None = None
        ) -> None:
            """Initialize the class."""
            if rules is None:
                Pseudonymize._Pseudonymizer.rules = []
            else:
                Pseudonymize._Pseudonymizer.rules.extend(rules)

            self.metadata = metadata

        def with_metadata(self, metadata: Datadoc) -> "Pseudonymize._Pseudonymizer":
            """Specify existing datadoc metadata for the dataset."""
            return Pseudonymize._Pseudonymizer(self.rules, metadata)

        def on_fields(self, *fields: str) -> "Pseudonymize._PseudoFuncSelector":
            """Specify one or multiple fields to be pseudonymized."""
            return Pseudonymize._PseudoFuncSelector(list(fields), self.metadata)

        def add_rules(
            self, rules: PseudoRule | list[PseudoRule]
        ) -> "Pseudonymize._Pseudonymizer":
            """Add one or more rules to existing pseudonymization rules."""
            if isinstance(rules, list):
                return Pseudonymize._Pseudonymizer(self.rules + rules, self.metadata)
            else:
                return Pseudonymize._Pseudonymizer([*self.rules, rules], self.metadata)

        def run(
            self,
            hierarchical: bool = False,
            custom_keyset: PseudoKeyset | str | None = None,
            timeout: int = TIMEOUT_DEFAULT,
        ) -> Result:
            """Pseudonymize the dataset.

            Args:
                hierarchical (bool): Whether the dataset is hierarchical or not. Needs PseudoRules with concrete paths. Defaults to False.
                custom_keyset (PseudoKeyset, optional): The pseudonymization keyset to use. Defaults to None.
                timeout (int): The timeout in seconds for the API call. Defaults to TIMEOUT_DEFAULT.

            Returns:
                Result: The pseudonymized dataset and the associated metadata.
            """
            super().__init__(
                pseudo_operation=PseudoOperation.PSEUDONYMIZE,
                dataset=Pseudonymize.dataset,
                hierarchical=hierarchical,
                user_provided_metadata=self.metadata,
            )

            result = super()._execute_pseudo_operation(
                self.rules, timeout, custom_keyset
            )
            return result

    class _PseudoFuncSelector(_BaseRuleConstructor):
        def __init__(self, fields: list[str], metadata: Datadoc | None) -> None:
            self._fields = fields
            self._metadata = metadata
            super().__init__(fields)

        def with_stable_id(
            self,
            sid_snapshot_date: str | date | None = None,
            custom_key: PredefinedKeys | str | None = None,
            on_map_failure: MapFailureStrategy | str | None = None,
        ) -> "Pseudonymize._Pseudonymizer":
            """Map the selected fields to Stable ID, then pseudonymize with a PAPIS-compatible encryption.

            In other words, this is a compound operation that both: 1) maps FNR to stable ID 2) then encrypts the Stable IDs.

            Args:
                sid_snapshot_date (Optional[str | date], optional): Date representing SID-catalogue version to use.
                    Latest if unspecified. Format: YYYY-MM-DD
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)
                on_map_failure (Optional[MapFailureStrategy | str], optional): defines how to handle mapping failures

            Returns:
                Self: The object configured to be mapped to stable ID
            """
            rules = super()._map_to_stable_id_and_pseudonymize(
                sid_snapshot_date, custom_key, on_map_failure
            )
            return Pseudonymize._Pseudonymizer(rules, self._metadata)

        def with_default_encryption(
            self, custom_key: PredefinedKeys | str | None = None
        ) -> "Pseudonymize._Pseudonymizer":
            """Pseudonymize the selected fields with the default encryption algorithm (DAEAD).

            Args:
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (ssb-common-key-1)

            Returns:
                Self: The object configured to be mapped to stable ID
            """
            rules = super()._with_daead_encryption(custom_key)
            return Pseudonymize._Pseudonymizer(rules, self._metadata)

        def with_papis_compatible_encryption(
            self, custom_key: PredefinedKeys | str | None = None
        ) -> "Pseudonymize._Pseudonymizer":
            """Pseudonymize the selected fields with a PAPIS-compatible encryption algorithm (FF31).

            Args:
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)

            Returns:
                Self: The object configured to be mapped to stable ID
            """
            rules = super()._with_ff31_encryption(custom_key)
            return Pseudonymize._Pseudonymizer(rules, self._metadata)

        def with_custom_function(
            self, function: PseudoFunction
        ) -> "Pseudonymize._Pseudonymizer":
            rules = super()._with_custom_function(function)
            return Pseudonymize._Pseudonymizer(rules, self._metadata)



import json
import typing as t
from datetime import date
from enum import Enum

from humps import camelize
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import FieldSerializationInfo
from pydantic import ValidationError
from pydantic import field_serializer
from pydantic import model_serializer
from pydantic import model_validator

from dapla_pseudo.constants import MapFailureStrategy
from dapla_pseudo.constants import PredefinedKeys
from dapla_pseudo.constants import PseudoFunctionTypes
from dapla_pseudo.constants import UnknownCharacterStrategy
from dapla_pseudo.models import APIModel


class Mimetypes(str, Enum):
    """Mimetypes is an enum of supported mimetypes, for use in HTTP requests.

    As a proxy, this also defines the supported input file formats when reading from a file.
    """

    JSON = "application/json"


class PseudoFunctionArgs(BaseModel):
    """Representation of the possible keyword arguments."""

    def __str__(self) -> str:
        """As a default, represent the fields of the subclasses as kwargs on the format 'k=v'."""
        return ",".join(
            f"{k}={v}"
            for k, v in self.model_dump(by_alias=True).items()
            if v is not None
        )

    model_config = ConfigDict(alias_generator=camelize, populate_by_name=True)


class MapSidKeywordArgs(PseudoFunctionArgs):
    """Representation of kwargs for the 'map-sid' function.

    Parameters:
        key_id: The key to be used for pseudonomization.
        snapshot_date (date): The timestamp for the version of the SID catalogue.
            If not specified, will choose the latest version.
            The format is: YYYY-MM-DD, e.g. 2021-05-21
        strategy: defines how encryption/decryption should handle non-alphabet characters
        failure_strategy: defines how to handle mapping failures
    """

    key_id: PredefinedKeys | str = Field(default=PredefinedKeys.PAPIS_COMMON_KEY_1)
    snapshot_date: date | None = None
    strategy: UnknownCharacterStrategy | None = UnknownCharacterStrategy.SKIP
    failure_strategy: MapFailureStrategy | None = None


class DaeadKeywordArgs(PseudoFunctionArgs):
    """Representation of kwargs for the 'daead' function."""

    key_id: PredefinedKeys | str = PredefinedKeys.SSB_COMMON_KEY_1


class FF31KeywordArgs(PseudoFunctionArgs):
    """Representation of kwargs for the 'FF31' function."""

    key_id: PredefinedKeys | str = PredefinedKeys.PAPIS_COMMON_KEY_1
    strategy: UnknownCharacterStrategy | None = UnknownCharacterStrategy.SKIP


class RedactKeywordArgs(PseudoFunctionArgs):
    """Representation of kwargs for the 'redact' function."""

    placeholder: str | None = None
    regex: str | None = None


class PseudoFunction(BaseModel):  # type: ignore[no-redef]
    """Formal representation of a pseudo function.

    Use to build up the string representation expected by pseudo service.

    Syntax: "<function_type>(<kwarg_1>=x, <kwarg_2>=y)"

    where <kwarg_1>, <kwarg_2>, etc. represents the keywords defined in PseudoFunctionArgs
    """

    function_type: PseudoFunctionTypes
    kwargs: DaeadKeywordArgs | FF31KeywordArgs | MapSidKeywordArgs | RedactKeywordArgs

    def __str__(self) -> str:
        """Create the function representation as expected by pseudo service."""
        return f"{self.function_type}({self.kwargs})"

    @model_serializer()
    def serialize_model(self) -> str:
        """Serialize the function as expected by the pseudo service."""
        return f"{self.function_type}({self.kwargs})"

    @model_validator(mode="before")
    @classmethod
    def deserialize_model(cls, data: str | dict[str, t.Any]) -> dict[str, t.Any]:
        """Deserialize the shorthand string representation of a pseudo function to Python model.

        This function parses the serialized version of a function like e.g. 'redact(placeholder=#)'
        by splitting the function name (fun) from the arguments (args), and then constructing a
        dict out of the args. Finally, the proper function type and kwargs are inferred from the
        PseudoFunctionTypes enum.
        """
        if isinstance(data, str):
            func: str
            args: str
            func, args = (
                data.replace(" ", "").replace("(", " ").replace(")", "").split(" ")
            )
            pseudo_function_type: PseudoFunctionTypes = PseudoFunctionTypes(func)
            args_dict: dict[str, str] | None = (
                dict(list(map(lambda v: v.split("="), args.split(","))))
                if len(args) > 0
                else None
            )
            return {
                "function_type": pseudo_function_type,
                "kwargs": cls._resolve_args(pseudo_function_type, args_dict),
            }
        else:
            return data

    @classmethod
    def _resolve_args(
        cls, pseudo_function_type: PseudoFunctionTypes, args: dict[str, str] | None
    ) -> DaeadKeywordArgs | FF31KeywordArgs | MapSidKeywordArgs | RedactKeywordArgs:
        match pseudo_function_type:
            case PseudoFunctionTypes.DAEAD:
                return (
                    DaeadKeywordArgs()
                    if args is None
                    else DaeadKeywordArgs.model_validate(args)
                )
            case PseudoFunctionTypes.REDACT:
                return (
                    RedactKeywordArgs()
                    if args is None
                    else RedactKeywordArgs.model_validate(args)
                )
            case PseudoFunctionTypes.FF31:
                return (
                    FF31KeywordArgs()
                    if args is None
                    else FF31KeywordArgs.model_validate(args)
                )
            case PseudoFunctionTypes.MAP_SID:
                return (
                    MapSidKeywordArgs()
                    if args is None
                    else MapSidKeywordArgs.model_validate(args)
                )


class PseudoRule(APIModel):
    """A ``PseudoRule`` defines a pattern, a transformation function, and optionally a friendly name of the rule.

    Each rule defines a glob pattern (see https://docs.oracle.com/javase/tutorial/essential/io/fileOps.html#glob)
    that identifies one or multiple fields, and a `func` that will be applied to the matching fields.

    Lists of PseudoRules are processed by the dapla-pseudo-service in the order they are defined, and only the first
    matching rule will be applied (thus: rule ordering is important).

    Parameters:
        pattern: Glob expression, such as: ``/**/{field1, field2, *navn}``
        func: A transformation function, such as ``tink-daead(<keyname>), redact(<replacementstring>) or fpe-anychar(<keyname>)``
        path: If hierarchical, ``/path/to/thing``, else None
        name: A friendly name of the rule. This is optional, but can be handy for debugging
    """

    pattern: str
    func: PseudoFunction
    path: str | None = None
    name: str | None = None

    @field_serializer("func")
    def serialize_func(
        self, func: PseudoFunction, _info: FieldSerializationInfo
    ) -> str:
        """Explicit serialization of the 'func' field to coerce to string before serializing PseudoRule."""
        return str(func)

    @classmethod
    def from_json(cls, data: str | dict[str, t.Any]) -> t.Any:
        """Deserialise the json-formatted pseudo rule to Python model."""
        if isinstance(data, str):
            return super().model_validate(json.loads(data))
        else:
            return super().model_validate(data)


class PseudoKeyset(APIModel):
    """PseudoKeyset represents a wrapped data encryption key (WDEK).

    Example structure, represented as JSON:
    {"encrypted_keyset": "CiQAp91NBhLdknX3j9jF6vwhdyURaqcT9/M/iczV7fLn...8XYFKwxiwMtCzDT6QGzCCCM=",
    "keyset_info": {
    "primaryKeyId": 1234567890,
    "keyInfo": [
    {
    "typeUrl": "type.googleapis.com/google.crypto.tink.AesSivKey",
    "status": "ENABLED",
    "keyId": 1234567890,
    "outputPrefixType": "TINK"
    }
    ]
    },
    "kek_uri": "gcp-kms://projects/some-project-id/locations/europe-north1/keyRings/some-keyring/cryptoKeys/some-kek-1"
    }
    """

    encrypted_keyset: str
    keyset_info: dict[str, t.Any]
    kek_uri: str

    def get_key_id(self) -> str:
        """ID of the keyset."""
        return str(self.keyset_info["primaryKeyId"])


class PseudoConfig(APIModel):
    """PseudoConfig is a container for rules and keysets."""

    rules: list[PseudoRule]
    keysets: list[PseudoKeyset] | None = None


class TargetCompression(APIModel):
    """TargetCompression denotes if and how results from the API should be compressed and password encrypted."""

    password: str


class KeyWrapper(BaseModel):
    """Hold information about a key, such as ID and keyset information."""

    key_id: str = ""
    keyset: PseudoKeyset | None = None

    def __init__(self, key: str | PseudoKeyset | None = None, **kwargs: t.Any) -> None:
        """Determine if a key is either a key reference (aka "common key") or a keyset.

        If it is a key reference, treat this as the key's ID, else retrieve the key's ID from the keyset data structure.

        :param key: either a key reference (as string) or a PseudoKeyset
        """
        super().__init__(**kwargs)
        if isinstance(key, str):
            try:  # Attempt to parse the key as a JSON-string matching the PseudoKeyset model
                pseudo_keyset = PseudoKeyset.model_validate(json.loads(key))
                self.key_id = pseudo_keyset.get_key_id()
                self.keyset = pseudo_keyset
                return
            except (ValidationError, json.JSONDecodeError):
                pass

            # Else, attempt to parse the key as one of the predefined keys
            if key in PredefinedKeys.__members__.values():
                self.key_id = key
                self.keyset = None
            else:
                raise ValueError(f"Key '{key}' is not a valid key reference or keyset")
        # Or we have an already parsed PseudoKeyset
        elif isinstance(key, PseudoKeyset):
            self.key_id = key.get_key_id()
            self.keyset = key

    def keyset_list(self) -> list[PseudoKeyset] | None:
        """Wrap the keyset in a list if it is defined - or return None if it is not."""
        return None if self.keyset is None else [self.keyset]



"""Type declarations for dapla-toolbelt-pseudo."""

import io
from pathlib import Path
from typing import TypeAlias

import fsspec
import gcsfs
import pandas as pd

FieldDecl: TypeAlias = str | dict[str, str]
BinaryFileDecl: TypeAlias = (
    io.BufferedReader
    | fsspec.spec.AbstractBufferedFile
    | gcsfs.core.GCSFile
    | io.BytesIO
)

DatasetDecl = pd.DataFrame | BinaryFileDecl | str | Path
FileLikeDatasetDecl = BinaryFileDecl | str | Path
FileSpecDecl = tuple[str | None, BinaryFileDecl | str, str]
# FileSpecDecl is derived from the "files" argument in multi-part requests from the "Requests"-library
# The tuple semantically means: ('filename', fileobj, 'content_type')
# See "files" in https://requests.readthedocs.io/en/latest/api/#requests.request



"""Common API models for builder packages."""

from collections import Counter
from pathlib import Path
from typing import Any
from typing import cast

import pandas as pd
import polars as pl
from cloudpathlib import GSClient
from cloudpathlib import GSPath
from dapla_metadata.datasets.core import Datadoc
from datadoc_model.all_optional.model import DatadocMetadata
from datadoc_model.all_optional.model import MetadataContainer
from datadoc_model.all_optional.model import Variable
from pydantic import ValidationError

from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.utils import get_file_format_from_file_name
from dapla_pseudo.v1.models.api import PseudoFieldResponse
from dapla_pseudo.v1.supported_file_format import write_from_df


class Result:
    """Result represents the result of a pseudonymization operation."""

    def __init__(
        self,
        pseudo_response: PseudoFieldResponse,
        pseudo_operation: PseudoOperation | None = None,
        targeted_columns: list[str] | None = None,
        user_provided_metadata: Datadoc | None = None,
    ) -> None:
        """Initialise a PseudonymizationResult."""
        self._pseudo_data: pl.DataFrame = pseudo_response.data
        self._metadata: dict[str, dict[str, list[Any]]] = {}
        self._datadoc: Datadoc | MetadataContainer

        datadoc_fields: list[Variable] = []
        datadoc_paths: list[str | None] = []

        for field_metadata in pseudo_response.raw_metadata:
            pseudo_variable: Variable | None = self._datadoc_from_raw_metadata_fields(
                field_metadata.datadoc
            )
            if (
                pseudo_variable is not None
                and pseudo_variable.data_element_path not in datadoc_paths
            ):
                datadoc_paths.append(pseudo_variable.data_element_path)
                datadoc_fields.append(pseudo_variable)

            if field_metadata.field_name is None:
                field_metadata.field_name = "unknown_field"

            # Add metadata per field
            self._metadata[field_metadata.field_name] = {
                "logs": field_metadata.logs,
                "metrics": field_metadata.metrics,
            }

        self._datadoc = self._construct_final_metadata(
            datadoc_fields,
            pseudo_operation,
            targeted_columns,
            user_provided_metadata,
        )

    def _construct_final_metadata(
        self,
        datadoc_fields: list[Variable],
        pseudo_operation: PseudoOperation | None,
        targeted_columns: list[str] | None,
        user_provided_metadata: Datadoc | None,
    ) -> Datadoc | MetadataContainer:
        """Construct the final datadoc metadata.

        If preexisting metadata is provided then modify that otherwise construct the Datadoc object.
        """
        datadoc_fields_map: dict[str, Variable] = {
            (
                datadoc_variable.short_name if datadoc_variable.short_name else ""
            ): datadoc_variable
            for datadoc_variable in datadoc_fields
        }
        if user_provided_metadata and targeted_columns and pseudo_operation:
            for column in targeted_columns:
                match pseudo_operation:
                    case PseudoOperation.PSEUDONYMIZE | PseudoOperation.REPSEUDONYMIZE:
                        user_provided_metadata.add_pseudonymization(
                            column,
                            datadoc_fields_map[column].pseudonymization,
                        )
                    case PseudoOperation.DEPSEUDONYMIZE:
                        user_provided_metadata.remove_pseudonymization(column)
            return user_provided_metadata
        else:
            return MetadataContainer(
                datadoc=DatadocMetadata(
                    document_version="5.0.1",
                    variables=datadoc_fields,
                )
            )

    def to_polars(self, **kwargs: Any) -> pl.DataFrame:
        """Output pseudonymized data as a Polars DataFrame.

        Args:
            **kwargs: Additional keyword arguments to be passed the Polars "from_dicts" function *if* the input data is from a file.

        Raises:
            ValueError: If the result is not of type Polars DataFrame.

        Returns:
            pl.DataFrame: A Polars DataFrame containing the pseudonymized data.
        """
        match self._pseudo_data:
            case pl.DataFrame() as df:
                # Drop statement a workaround to https://github.com/pola-rs/polars/issues/7291
                if "__index_level_0__" in df.columns:
                    df = df.drop("__index_level_0__")
                return df
            case _ as invalid_pseudo_data:
                raise ValueError(f"Invalid file type: {type(invalid_pseudo_data)}")

    def to_pandas(self, **kwargs: Any) -> pd.DataFrame:
        """Output pseudonymized data as a Pandas DataFrame.

        Args:
            **kwargs: Additional keyword arguments to be passed the Pandas reader function *if* the input data is from a file.
                The specific reader function depends on the format of the input file, e.g. `read_csv()` for CSV files.

        Raises:
            ValueError: If the result is not of type Polars DataFrame.

        Returns:
            pd.DataFrame: A Pandas DataFrame containing the pseudonymized data.
        """
        match self._pseudo_data:
            case pl.DataFrame() as df:
                return df.to_pandas()
            case _ as invalid_pseudo_data:
                raise ValueError(f"Invalid response type: {type(invalid_pseudo_data)}")

    def to_file(self, file_path: str, **kwargs: Any) -> None:
        """Write pseudonymized data to a file, with the metadata being written to the same folder.

        Args:
            file_path (str): The path to the file to be written. If writing to a bucket, use the "gs://" prefix.
            **kwargs: Additional keyword arguments to be passed the Polars writer function *if* the input data is a DataFrame.
                The specific writer function depends on the format of the output file, e.g. `write_csv()` for CSV files.

        Raises:
            ValueError: If the result is not of type Polars DataFrame.
            ValueError: If the output file format does not match the input file format.

        """
        file_format = get_file_format_from_file_name(file_path)

        datadoc_file_name = f"{Path(file_path).stem}__DOC.json"

        datadoc_file_path: Path | GSPath
        if file_path.startswith(GSPath.cloud_prefix):
            client = GSClient()
            gs_path = GSPath(file_path, client)

            file_handle = gs_path.open(mode="wb")

            datadoc_file_path = gs_path.parent.joinpath(Path(datadoc_file_name))
            datadoc_file_handle = datadoc_file_path.open(mode="w")
        else:
            file_handle = Path(file_path).open(mode="wb")

            datadoc_file_path = Path(file_path).parent.joinpath(Path(datadoc_file_name))
            datadoc_file_handle = datadoc_file_path.open(mode="w")

        match self._pseudo_data:
            case pl.DataFrame() as df:
                write_from_df(df, file_format, file_handle, **kwargs)
                datadoc_file_handle.write(self.datadoc)
            case _ as invalid_pseudo_data:
                raise ValueError(f"Invalid response type: {type(invalid_pseudo_data)}")

        file_handle.close()
        datadoc_file_handle.close()

    @property
    def metadata_details(self) -> dict[str, Any]:
        """Returns the pseudonymization metadata as a dictionary, for each field that has been processed.

        Returns:
            Optional[dict[str, str]]: A dictionary containing the pseudonymization metadata,
            where the keys are field names and the values are corresponding pseudo field metadata.
            If no metadata is set, returns an empty dictionary.
        """
        return self._metadata

    @property
    def metadata(self) -> dict[str, Any]:
        """Returns the aggregated metadata for all fields as a dictionary.

        Returns:
            Optional[dict[str, str]]: A dictionary containing the pseudonymization metadata,
            where the keys are field names and the values are corresponding pseudo field metadata.
            If no metadata is set, returns an empty dictionary.
        """
        return aggregate_metrics(self._metadata)

    @property
    def datadoc_model(self) -> dict[str, Any]:
        """Returns the pseudonymization metadata as a dictionary.

        Returns:
            dict: A dictionary representing the datadoc metadata.
        """
        match self._datadoc:
            case Datadoc():
                return self._datadoc.datadoc_model().model_dump(exclude_none=True)
            case MetadataContainer():
                return self._datadoc.model_dump(exclude_none=True)

    @property
    def datadoc(self) -> str:
        """Returns the pseudonymization metadata as a string.

        Returns:
            str: A JSON-formattted string representing the datadoc metadata.
        """
        match self._datadoc:
            case Datadoc():
                return self._datadoc.datadoc_model().model_dump_json(exclude_none=True)
            case MetadataContainer():
                return self._datadoc.model_dump_json(exclude_none=True)

    def _datadoc_from_raw_metadata_fields(
        self,
        raw_metadata: list[dict[str, Any]] | None,
    ) -> Variable | None:

        # This happens if we recieve old datadoc metadata from pseudo-service
        if raw_metadata is None:
            print(
                """[WARNING] 'variables' key not found in datadoc metadata recieved from pseudo-service. This means the metadata is from an outdated datadoc model."""
            )
            return None
        elif len(raw_metadata) == 0:
            return None
        elif len(raw_metadata) > 1:
            print(f"Unexpected length of metadata: {len(raw_metadata)}")

        try:
            variable = Variable(**raw_metadata[0])
            return Variable.model_validate(variable)
        except ValidationError as exc:
            print(
                f"""[WARNING] Datadoc 'Variable' validation failed:
                {exc}"""
            )
            return None


def aggregate_metrics(metadata: dict[str, dict[str, list[Any]]]) -> dict[str, Any]:
    """Aggregates logs and metrics. Each unique metric is summarized."""
    logs: list[str] = []
    metrics: Counter[str] = Counter()
    for field_metadata in metadata.values():
        for metadata_field, metadata_value in field_metadata.items():
            match metadata_field:
                case "logs":
                    # Logs are simply appended
                    logs.extend(metadata_value)
                case "metrics":
                    # Count each unique metric
                    metadata_value = cast(list[dict[str, int]], metadata_value)
                    # Metrics is represented by a list of dicts, e.g. [{"METRIC_1": 1}, {"METRIC_2": 1}]
                    for metric in metadata_value:
                        # Each dict has a single entry, e.g. {"METRIC_1": 1}, so take the first one
                        key, value = next(iter(metric.items()))
                        metrics.update({key: value})
    return {"logs": logs, "metrics": dict(metrics)}



"""Module that implements a client abstraction that makes it easy to communicate with the Dapla Pseudo Service REST API."""

import asyncio
import copy
import os
import typing as t
from collections import defaultdict
from collections.abc import Generator
from datetime import date

import google.auth.transport.requests
import google.oauth2.id_token
import requests
from aiohttp import ClientResponse
from aiohttp import ClientSession
from aiohttp import ClientTimeout
from aiohttp import TCPConnector
from aiohttp_retry import ExponentialRetry
from aiohttp_retry import RetryClient
from dapla_auth_client import AuthClient
from ulid import ULID

from dapla_pseudo.constants import TIMEOUT_DEFAULT
from dapla_pseudo.constants import Env
from dapla_pseudo.constants import PseudoFunctionTypes
from dapla_pseudo.utils import redact_field
from dapla_pseudo.v1.models.api import DepseudoFieldRequest
from dapla_pseudo.v1.models.api import PseudoFieldRequest
from dapla_pseudo.v1.models.api import RawPseudoMetadata
from dapla_pseudo.v1.models.api import RepseudoFieldRequest
from dapla_pseudo.v1.models.core import Mimetypes


class PseudoClient:
    """Client for interacting with the Dapla Pseudo Service REST API."""

    def __init__(
        self,
        pseudo_service_url: str | None = None,
        auth_token: str | None = None,
        rows_per_partition: str | None = None,
        max_total_partitions: str | None = None,
    ) -> None:
        """Use a default url for dapla-pseudo-service if not explicitly set."""
        self.pseudo_service_url = (
            "http://dapla-pseudo-service.dapla.svc.cluster.local"
            if pseudo_service_url is None
            else pseudo_service_url
        )
        self.static_auth_token = auth_token
        self.rows_per_partition = (
            10000 if rows_per_partition is None else int(rows_per_partition)
        )
        self.max_total_partitions = (
            200 if max_total_partitions is None else int(max_total_partitions)
        )

    def __auth_token(self) -> str:
        if os.environ.get("DAPLA_REGION") == "CLOUD_RUN":
            audience = os.environ["PSEUDO_SERVICE_URL"]
            auth_req = google.auth.transport.requests.Request()  # type: ignore[no-untyped-call]
            token = t.cast(
                str,
                google.oauth2.id_token.fetch_id_token(auth_req, audience),  # type: ignore[no-untyped-call]
            )
            return token
        else:
            return (
                str(AuthClient.fetch_personal_token())
                if self.static_auth_token is None
                else str(self.static_auth_token)
            )

    async def post_to_field_endpoint(
        self,
        path: str,
        timeout: int,
        pseudo_requests: list[
            PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest
        ],
    ) -> list[tuple[str, list[str | None], RawPseudoMetadata]]:
        """Post a request to the Pseudo Service field endpoint.

        Args:
            path (str): Full URL to the endpoint
            timeout (int): Request timeout
            pseudo_requests: Pseudo requests

        Returns:
            list[tuple[str, list[str], RawPseudoMetadata]]: A list of tuple of (field_name, data, metadata)
        """

        async def _post(
            client: RetryClient,
            path: str,
            timeout: int,
            correlation_id: str,
            request: PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest,
        ) -> tuple[str, list[str | None], RawPseudoMetadata]:
            if (
                type(request) is PseudoFieldRequest
                and request.pseudo_func.function_type == PseudoFunctionTypes.REDACT
            ):
                return redact_field(request)
            else:
                async with client.post(
                    url=f"{self.pseudo_service_url}/{path}",
                    headers={
                        "Authorization": f"Bearer {self.__auth_token()}",
                        "Content-Type": Mimetypes.JSON.value,
                        "X-Correlation-Id": correlation_id,
                    },
                    json={"request": request.model_dump(by_alias=True)},
                    timeout=timeout,
                ) as response:
                    await PseudoClient._handle_response_error(response)
                    response_json = await response.json()
                    data = response_json["data"]
                    metadata = RawPseudoMetadata(
                        field_name=request.name,
                        logs=response_json["logs"],
                        metrics=response_json["metrics"],
                        datadoc=response_json["datadoc_metadata"].get(
                            "variables", None
                        ),
                    )

                    return request.name, data, metadata

        split_pseudo_requests = self._split_requests(pseudo_requests)
        aio_session = ClientSession(
            connector=TCPConnector(limit=200, force_close=True),
            timeout=ClientTimeout(total=TIMEOUT_DEFAULT),
        )
        async with RetryClient(
            client_session=aio_session,
            retry_options=ExponentialRetry(
                attempts=5,
                start_timeout=0.1,
                max_timeout=30,
                factor=6,
                statuses={
                    400,
                }.union(
                    set(range(500, 600))
                ),  # Retry all 5xx errors and 400 Bad Request
            ),
        ) as client:
            results = await asyncio.gather(
                *[
                    _post(
                        client=client,
                        path=path,
                        timeout=timeout,
                        request=req,
                        correlation_id=correlation_id,
                    )
                    for (correlation_id, reqs) in split_pseudo_requests.items()
                    for req in reqs
                ]
            )
        await asyncio.sleep(0.1)  # Allow time for sockets to close
        await aio_session.close()

        return PseudoClient._merge_responses(results)

    def _split_requests(
        self,
        field_requests: list[
            PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest
        ],
    ) -> dict[
        str, list[PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest]
    ]:
        """Split requests into partitions.

        This is done to limit the size of a single request and more evenly distribute the load across the Pseudo Service.
        """

        def partition_requests(
            field_request: (
                PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest
            ),
            chunk_size: int,
        ) -> Generator[
            PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest,
            None,
            None,
        ]:
            for i in range(0, len(field_request.values), chunk_size):
                new_field_request = copy.deepcopy(field_request)
                new_field_request.values = field_request.values[i : i + chunk_size]
                yield new_field_request

        partitioned_field_requests: dict[
            str, list[PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest]
        ] = defaultdict(list)
        for req in field_requests:
            n_partitions = min(
                self.max_total_partitions // len(field_requests),
                len(req.values) // (self.rows_per_partition * len(field_requests)),
            )
            # We delegate equal amount of partitions to every field.

            if n_partitions <= 1:  # Do not split if number of rows is less than 10k
                partitioned_field_requests[
                    PseudoClient._generate_new_correlation_id()
                ].append(req)
                continue

            correlation_id = PseudoClient._generate_new_correlation_id()
            partition_size = max(1, len(req.values) // n_partitions)
            for request in partition_requests(req, partition_size):
                partitioned_field_requests[correlation_id].append(request)

        return partitioned_field_requests

    @staticmethod
    def _merge_responses(
        responses: list[tuple[str, list[str | None], RawPseudoMetadata]],
    ) -> list[tuple[str, list[str | None], RawPseudoMetadata]]:
        """Merge the response from the Pseudo Service into a single tuple.

        The responses are merged such that the first value of tuple, the field name, is unique.
        The second value, the pseudonymized data, gets concatenated with the other field names.
        The third value, the metadata - for Datadoc metadata they should always be equal, so we just keep the first one encountered.
        For logs and metrics, we concatenate the lists/dictionaries of lists.
        """
        grouped: dict[str, tuple[list[str | None], RawPseudoMetadata]] = defaultdict(
            lambda: ([], RawPseudoMetadata(logs=[], metrics=[], datadoc=None))
        )

        for key, string_list, metadata in responses:
            grouped[key] = (grouped[key][0] + (string_list), grouped[key][1] + metadata)

        return [(k, v[0], v[1]) for k, v in grouped.items()]

    @staticmethod
    def _generate_new_correlation_id() -> str:
        return str(ULID())

    @staticmethod
    async def _handle_response_error(response: ClientResponse) -> None:
        """Report error messages in response object."""
        match response.status:
            case status if status in range(200, 300):
                pass
            case _:
                print(response.headers)
                print(await response.text())
                response.raise_for_status()

    @staticmethod
    def _handle_response_error_sync(response: requests.Response) -> None:
        """Report error messages in response object. For synchronous callers."""
        match response.status_code:
            case status if status in range(200, 300):
                pass
            case _:
                print(response.headers)
                print(response.text)
                response.raise_for_status()

    def _post_to_sid_endpoint(
        self,
        path: str,
        values: list[str],
        sid_snapshot_date: date | None = None,
        stream: bool = True,
    ) -> requests.Response:
        request: dict[str, t.Collection[str]] = {"fnrList": values}
        response = requests.post(
            url=f"{self.pseudo_service_url}/{path}",
            params={"snapshot": str(sid_snapshot_date)} if sid_snapshot_date else None,
            # Do not set content-type, as this will cause the json to serialize incorrectly
            headers={
                "Authorization": f"Bearer {self.__auth_token()}",
                "X-Correlation-Id": PseudoClient._generate_new_correlation_id(),
            },
            json=request,
            stream=stream,
            timeout=TIMEOUT_DEFAULT,  # seconds
        )

        PseudoClient._handle_response_error_sync(response)
        return response


def _client() -> PseudoClient:
    return PseudoClient(
        pseudo_service_url=os.getenv(Env.PSEUDO_SERVICE_URL),
        auth_token=os.getenv(Env.PSEUDO_SERVICE_AUTH_TOKEN),
    )



"""Baseclasses for the Pseudo Builders.

The methods are kept private in order to not expose them to users of the client
when using autocomplete-features. The method names should also be more technical
and descriptive than the user-friendly methods that are exposed.
"""

import asyncio
import os
from concurrent.futures import ThreadPoolExecutor
from datetime import date

import polars as pl
from dapla_metadata.datasets.core import Datadoc

from dapla_pseudo.constants import Env
from dapla_pseudo.constants import MapFailureStrategy
from dapla_pseudo.constants import PredefinedKeys
from dapla_pseudo.constants import PseudoFunctionTypes
from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.utils import build_pseudo_field_request
from dapla_pseudo.utils import convert_to_date
from dapla_pseudo.utils import running_asyncio_loop
from dapla_pseudo.v1.client import PseudoClient
from dapla_pseudo.v1.models.api import DepseudoFieldRequest
from dapla_pseudo.v1.models.api import PseudoFieldRequest
from dapla_pseudo.v1.models.api import PseudoFieldResponse
from dapla_pseudo.v1.models.api import RawPseudoMetadata
from dapla_pseudo.v1.models.api import RepseudoFieldRequest
from dapla_pseudo.v1.models.core import DaeadKeywordArgs
from dapla_pseudo.v1.models.core import FF31KeywordArgs
from dapla_pseudo.v1.models.core import MapSidKeywordArgs
from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoKeyset
from dapla_pseudo.v1.models.core import PseudoRule
from dapla_pseudo.v1.mutable_dataframe import MutableDataFrame
from dapla_pseudo.v1.result import Result


class _BasePseudonymizer:
    """Base class for the _Pseudonymizer/_Depseudonymizer/_Repseudonymizer builders."""

    def __init__(
        self,
        pseudo_operation: PseudoOperation,
        dataset: pl.DataFrame,
        hierarchical: bool,
        user_provided_metadata: Datadoc | None,
    ) -> None:
        """The constructor of the base class."""
        self._pseudo_operation = pseudo_operation
        self._pseudo_client: PseudoClient = PseudoClient(
            pseudo_service_url=os.getenv(Env.PSEUDO_SERVICE_URL),
            auth_token=os.getenv(Env.PSEUDO_SERVICE_AUTH_TOKEN),
            rows_per_partition=os.getenv(Env.PSEUDO_CLIENT_ROWS_PER_PARTITION),
            max_total_partitions=os.getenv(Env.PSEUDO_CLIENT_MAX_TOTAL_PARTITIONS),
        )
        self._dataset = MutableDataFrame(dataset, hierarchical)
        self._user_provided_metadata = user_provided_metadata

    def _execute_pseudo_operation(
        self,
        rules: list[PseudoRule],  # "source rules" if repseudo
        timeout: int,
        custom_keyset: PseudoKeyset | str | None = None,
        target_custom_keyset: PseudoKeyset | str | None = None,  # used in repseudo
        target_rules: list[PseudoRule] | None = None,  # used in repseudo
    ) -> Result:
        if self._dataset is None:
            raise ValueError("No dataset has been provided.")

        if rules == []:
            raise ValueError(
                "No fields have been provided. Use the 'on_fields' method."
            )

        pseudo_requests = build_pseudo_field_request(
            self._pseudo_operation,
            self._dataset,
            rules,
            custom_keyset,
            target_custom_keyset,
            target_rules,
        )

        pseudo_response = self._pseudonymize_field(pseudo_requests, timeout)
        return Result(
            pseudo_response=pseudo_response,
            pseudo_operation=self._pseudo_operation,
            targeted_columns=[
                pseudo_rule.pattern
                for pseudo_rule in (target_rules if target_rules else rules)
            ],
            user_provided_metadata=self._user_provided_metadata,
        )

    def _pseudonymize_field(
        self,
        pseudo_requests: list[
            PseudoFieldRequest | DepseudoFieldRequest | RepseudoFieldRequest
        ],
        timeout: int,
    ) -> PseudoFieldResponse:
        """Pseudonymizes the specified fields in the DataFrame using the provided pseudonymization function.

        The pseudonymization is performed concurrently. After the processing is finished,
        the pseudonymized fields replace the original fields in the DataFrame stored in `self._dataframe`.
        """
        # type narrowing isn't carried over from caller function
        assert isinstance(self._dataset, MutableDataFrame)
        # Execute the pseudonymization API calls in parallel

        raw_metadata_fields: list[RawPseudoMetadata] = []
        if running_asyncio_loop() is not None:
            with ThreadPoolExecutor(
                1
            ) as pool:  # Run new event loop in a second worker thread if an event loop is already running
                result = pool.submit(
                    lambda: asyncio.run(
                        self._pseudo_client.post_to_field_endpoint(
                            path=f"{self._pseudo_operation.value}/field",
                            timeout=timeout,
                            pseudo_requests=pseudo_requests,
                        )
                    )
                ).result()
        else:
            result = asyncio.run(
                self._pseudo_client.post_to_field_endpoint(
                    path=f"{self._pseudo_operation.value}/field",
                    timeout=timeout,
                    pseudo_requests=pseudo_requests,
                )
            )

        for field_name, data, raw_metadata in result:
            self._dataset.update(field_name, data)
            raw_metadata_fields.append(raw_metadata)

        return PseudoFieldResponse(
            data=self._dataset.to_polars(), raw_metadata=raw_metadata_fields
        )


class _BaseRuleConstructor:
    """Base class for the _PseudoFuncSelector/_DepseudoFuncSelector/_RepseudoFuncSelector builders."""

    def __init__(
        self,
        fields: list[str],
    ) -> None:
        self._fields = fields

    def _map_to_stable_id_and_pseudonymize(
        self,
        sid_snapshot_date: str | date | None = None,
        custom_key: PredefinedKeys | str | None = None,
        on_map_failure: MapFailureStrategy | str | None = None,
    ) -> list[PseudoRule]:
        failure_strategy = (
            None if on_map_failure is None else MapFailureStrategy(on_map_failure)
        )
        kwargs = (
            MapSidKeywordArgs(
                key_id=custom_key,
                snapshot_date=convert_to_date(sid_snapshot_date),
                failure_strategy=failure_strategy,
            )
            if custom_key
            else MapSidKeywordArgs(
                snapshot_date=convert_to_date(sid_snapshot_date),
                failure_strategy=failure_strategy,
            )
        )
        pseudo_func = PseudoFunction(
            function_type=PseudoFunctionTypes.MAP_SID, kwargs=kwargs
        )
        return self._rule_constructor(pseudo_func)

    def _with_daead_encryption(
        self, custom_key: PredefinedKeys | str | None = None
    ) -> list[PseudoRule]:
        kwargs = (
            DaeadKeywordArgs(key_id=custom_key) if custom_key else DaeadKeywordArgs()
        )
        pseudo_func = PseudoFunction(
            function_type=PseudoFunctionTypes.DAEAD, kwargs=kwargs
        )
        return self._rule_constructor(pseudo_func)

    def _with_ff31_encryption(
        self, custom_key: PredefinedKeys | str | None = None
    ) -> list[PseudoRule]:
        kwargs = FF31KeywordArgs(key_id=custom_key) if custom_key else FF31KeywordArgs()
        pseudo_func = PseudoFunction(
            function_type=PseudoFunctionTypes.FF31, kwargs=kwargs
        )
        return self._rule_constructor(pseudo_func)

    def _with_custom_function(self, function: PseudoFunction) -> list[PseudoRule]:
        return self._rule_constructor(function)

    def _rule_constructor(self, func: PseudoFunction) -> list[PseudoRule]:
        # If we use the pseudonymize_file endpoint, we need a glob catch-all prefix.
        rules = [
            PseudoRule(name=None, func=func, pattern=field) for field in self._fields
        ]
        return rules



"""Classes used to support reading of dataframes from file."""

import json
import typing as t
from enum import Enum
from io import BufferedWriter
from io import BytesIO
from pathlib import Path
from typing import Any

import pandas as pd
import polars as pl

from dapla_pseudo.exceptions import ExtensionNotValidError


class SupportedOutputFileFormat(Enum):
    """SupportedOutputFileFormat contains the supported file formats when outputting the result to a file.

    Note that this does NOT describe the valid file extensions of _input_ data when reading from a file.
    """

    ZIP = "zip"
    CSV = "csv"
    JSON = "json"
    XML = "xml"
    PARQUET = "parquet"

    @classmethod
    def _missing_(cls, value: object) -> None:
        raise ExtensionNotValidError(
            f"{value} is not a valid file format. Valid formats: %s"
            % (", ".join([repr(m.value) for m in cls]))
        )


FORMAT_TO_MIMETYPE_FUNCTION = {
    SupportedOutputFileFormat.CSV: "text/csv",
    SupportedOutputFileFormat.JSON: "application/json",
    SupportedOutputFileFormat.ZIP: "application/zip",
}


def read_to_pandas_df(
    supported_format: SupportedOutputFileFormat,
    df_dataset: BytesIO | Path,
    **kwargs: Any,
) -> pd.DataFrame:
    """Reads a file with a supported file format to a Pandas Dataframe."""
    match supported_format:
        case SupportedOutputFileFormat.CSV:
            return pd.DataFrame(
                pd.read_csv(df_dataset, sep=";", **kwargs)
            )  # Pseudo Service CSV-separator is ';'
        case SupportedOutputFileFormat.JSON:
            return pd.DataFrame(pd.read_json(df_dataset, **kwargs))
        case SupportedOutputFileFormat.XML:
            return pd.read_xml(df_dataset, **kwargs)
        case SupportedOutputFileFormat.PARQUET:
            return pd.read_parquet(df_dataset, **kwargs)
        case SupportedOutputFileFormat.ZIP:
            raise ValueError(
                f"Unsupported file format for Pandas: '{supported_format}'."
            )


def read_to_polars_df(
    supported_format: SupportedOutputFileFormat,
    df_dataset: BytesIO | Path,
    **kwargs: Any,
) -> pl.DataFrame:
    """Reads a file with a supported file format to a Polars Dataframe."""
    match supported_format:
        case SupportedOutputFileFormat.CSV:
            return pl.read_csv(
                df_dataset, separator=";", **kwargs
            )  # Pseudo Service CSV-separator is ';'
        case SupportedOutputFileFormat.JSON:
            return pl.read_json(df_dataset, **kwargs)
        case SupportedOutputFileFormat.PARQUET:
            return pl.read_parquet(df_dataset, **kwargs)
        case SupportedOutputFileFormat.XML | SupportedOutputFileFormat.ZIP:
            raise ValueError(
                f"Unsupported file format for Polars: '{supported_format}'."
            )


def write_from_dicts(
    data: list[dict[str, t.Any]],
    supported_format: SupportedOutputFileFormat,
    file_like: BufferedWriter,
) -> None:
    """Writes data from a list of dicts to a file of the given format."""
    match supported_format:
        case SupportedOutputFileFormat.PARQUET:
            df = pl.DataFrame(data)
            # type hints lying
            df.write_parquet(file_like)  # type: ignore[arg-type]
        case SupportedOutputFileFormat.CSV:
            df = pl.DataFrame(data)
            df.write_csv(file_like)
        case SupportedOutputFileFormat.JSON:
            file_like.write(bytes(json.dumps(data), encoding="utf-8"))
        case SupportedOutputFileFormat.XML:
            df_pandas = pd.DataFrame.from_records(data)
            df_pandas.to_xml(file_like)
        case _:
            raise ValueError("Unsupported output file format")


def write_from_df(
    df: pl.DataFrame,
    supported_format: SupportedOutputFileFormat,
    file_like: BufferedWriter,
    **kwargs: Any,
) -> None:
    """Writes to a file with a supported file format from a Dataframe."""
    match supported_format:
        case SupportedOutputFileFormat.CSV:
            df.write_csv(file=file_like, **kwargs)
        case SupportedOutputFileFormat.JSON:
            df.write_json(file=file_like, **kwargs)
        case SupportedOutputFileFormat.XML:
            df.to_pandas().to_xml(file_like, **kwargs)
        case SupportedOutputFileFormat.PARQUET:
            df.write_parquet(file_like, **kwargs)  # type: ignore[arg-type]



"""This module defines helper classes and API models used to communicate with the Dapla Pseudo Service."""

import typing as t
from collections import defaultdict
from dataclasses import dataclass

import polars as pl

from dapla_pseudo.models import APIModel
from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoKeyset


class PseudoFieldRequest(APIModel):
    """Model of the pseudo field request sent to the service."""

    pseudo_func: PseudoFunction
    name: str
    pattern: str
    values: list[str | int | None]  # 'int' is necessary to run redact on integers
    keyset: PseudoKeyset | None = None


class DepseudoFieldRequest(APIModel):
    """Model of the depseudo field request sent to the service."""

    pseudo_func: PseudoFunction
    name: str
    pattern: str
    values: list[str | int | None]
    keyset: PseudoKeyset | None = None


class RepseudoFieldRequest(APIModel):
    """Model of the repseudo field request sent to the service."""

    source_pseudo_func: PseudoFunction
    target_pseudo_func: PseudoFunction | None
    name: str
    pattern: str
    values: list[str | int | None]
    source_keyset: PseudoKeyset | None = None
    target_keyset: PseudoKeyset | None = None


@dataclass
class RawPseudoMetadata:
    """RawPseudoMetadata holds the raw metadata obtained from Pseudo Service."""

    logs: list[str]
    metrics: list[dict[str, int]]
    datadoc: list[dict[str, t.Any]] | None
    field_name: str | None = None

    def __add__(self, other: "RawPseudoMetadata | None") -> "RawPseudoMetadata":
        """Combine two RawPseudoMetadata instances."""
        if other is None:
            return self
        grouped_metrics: dict[str, int] = defaultdict(int)
        for item in self.metrics + other.metrics:
            for key, value in item.items():
                grouped_metrics[key] += value
        grouped_metrics_unrolled = [{k: v} for k, v in grouped_metrics.items()]

        return RawPseudoMetadata(
            logs=self.logs + other.logs,
            metrics=grouped_metrics_unrolled,
            datadoc=self.datadoc or other.datadoc,
            field_name=self.field_name,
        )


@dataclass
class PseudoFieldResponse:
    """PseudoFieldResponse holds the data and metadata from a Pseudo Service field response."""

    data: pl.DataFrame
    raw_metadata: list[RawPseudoMetadata]



import re
from collections import Counter
from collections.abc import Generator
from io import BytesIO
from typing import Any

import msgspec
import orjson
import polars as pl

from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoRule

ARRAY_INDEX_MATCHER = re.compile(r"\[\d*]")


def _ensure_normalized(pattern: str) -> str:
    """Normalize the pattern.

    Ensure that the pattern always starts with a '/' or '*' to be compatible with the
    pattern matching that is used in pseudo-service.
    """
    return (
        pattern
        if (pattern.startswith("/") or pattern.startswith("*"))
        else "/" + pattern
    )


class FieldMatch:
    """Represents a reference to a matching column in the dataframe."""

    def __init__(
        self,
        path: str,
        pattern: str,
        col: dict[str, Any] | list[str | int | None],
        func: PseudoFunction,  # "source func" if repseudo
        target_func: PseudoFunction | None,  # "target_func" if repseudo
    ) -> None:
        """Initialize the class."""
        self.path = path
        self.pattern = _ensure_normalized(pattern)
        self.col = col
        self.func = func
        self.target_func = target_func

    def get_value(self) -> list[str | int | None]:
        """Get the inner value.

        If hierarchical, get the values of the matched column.
        Otherwise, just return the data of the Polars DataFrame.
        """
        if isinstance(self.col, list):
            return self.col
        else:
            return self.col["values"]  # type: ignore[no-any-return]


class MutableDataFrame:
    """A DataFrame that can change values in-place.

    If the DataFrame is hierarchical
    """

    def __init__(self, dataframe: pl.DataFrame, hierarchical: bool) -> None:
        """Initialize the class."""
        self.dataset: pl.DataFrame | dict[str, Any] = dataframe
        self.matched_fields: dict[str, FieldMatch] = {}
        self.matched_fields_metrics: dict[str, int] | None = None
        self.hierarchical: bool = hierarchical
        self.schema = dataframe.schema

    def match_rules(
        self, rules: list[PseudoRule], target_rules: list[PseudoRule] | None
    ) -> None:
        """Create references to all the columns that matches the given pseudo rules."""
        if self.hierarchical is False:
            assert isinstance(self.dataset, pl.DataFrame)
            self.matched_fields = {
                str(i): FieldMatch(
                    path=rule.pattern,
                    pattern=rule.pattern,
                    col=list(self.dataset.get_column(rule.pattern)),
                    func=rule.func,
                    target_func=target_rule.func if target_rule else None,
                )
                for (i, (rule, target_rule)) in enumerate(
                    _combine_rules(rules, target_rules)
                )
            }
        else:
            counter: Counter[str] = Counter()
            assert isinstance(self.dataset, pl.DataFrame)
            self.dataset = msgspec.json.decode(self.dataset.serialize(format="json"))
            assert isinstance(self.dataset, dict)
            for source_rule, target_rule in _combine_rules(rules, target_rules):
                if source_rule.path is None:
                    raise ValueError(
                        f"Rule: {source_rule}\n does not have a concrete path, and cannot be used."
                    )
                matches = _traverse_dataframe_dict(
                    self.dataset["columns"],
                    (source_rule, target_rule),
                    source_rule.path.split("/"),
                    counter,
                )
                for match in matches:
                    self.matched_fields[match.path] = match

            # The Counter contains unique field names. A count > 1 means that the traverse
            # was not able to group all values with a given path. This will be the case for
            # list of dicts.
            self.matched_fields_metrics = dict(counter)

    def get_matched_fields(self) -> dict[str, FieldMatch]:
        """Get a reference to all the columns that matched pseudo rules."""
        return self.matched_fields

    def update(self, path: str, data: list[str | None]) -> None:
        """Update a column with the given data."""
        if self.hierarchical is False:
            assert isinstance(self.dataset, pl.DataFrame)
            self.dataset = self.dataset.with_columns(pl.Series(data).alias(path))
        elif (field_match := self.matched_fields.get(path)) is not None:
            assert isinstance(field_match.col, dict)
            field_match.col.update({"values": data})

    def to_polars(self) -> pl.DataFrame:
        """Convert to Polars DataFrame."""
        if self.hierarchical is False:
            assert isinstance(self.dataset, pl.DataFrame)
            return self.dataset
        else:
            return pl.DataFrame.deserialize(
                BytesIO(orjson.dumps(self.dataset)),
                format="json",
            )


def _combine_rules(
    rules: list[PseudoRule], target_rules: list[PseudoRule] | None
) -> list[tuple[PseudoRule, PseudoRule | None]]:
    combined: list[tuple[PseudoRule, PseudoRule | None]] = []

    # Zip rules and target_rules together; use None as target if target_rules is undefined
    for index, rule in enumerate(rules):
        combined.append((rule, target_rules[index] if target_rules else None))

    return combined


def _traverse_dataframe_dict(
    items: list[dict[str, Any] | None],
    rules: tuple[PseudoRule, PseudoRule | None],
    curr_path: list[str],
    metrics: Counter[str],
    prefix: str = "",
) -> Generator[FieldMatch, None, None]:

    path_head, *path_tail = curr_path
    for index, col in enumerate(items):
        if col is None or curr_path == []:
            continue

        if col["name"] == "" and any(
            key in col["datatype"] for key in {"Struct", "List", "Array"}
        ):
            next_prefix = f"{prefix}[{index}]"
            yield from _traverse_dataframe_dict(
                col["values"],
                rules=rules,
                curr_path=curr_path,
                metrics=metrics,
                prefix=next_prefix,
            )
        elif col["name"] == path_head:
            next_prefix = f"{prefix}/{col['name']}"
            if path_tail == []:  # matched entire path
                rule, target_rule = rules

                if (
                    "List" in col["datatype"] or "Array" in col["datatype"]
                ):  # is pl.List or pl.Array
                    ## Special case: inner lists are weird and needs to be wrangled.
                    col = col["values"][0]
                    if col is None or len(col) == 0:
                        continue

                if not all(v is None for v in col["values"]):
                    metrics.update({next_prefix: 1})
                    yield FieldMatch(
                        path=next_prefix.lstrip("/"),
                        col=col,
                        func=rule.func,
                        target_func=target_rule.func if target_rule else None,
                        pattern=rule.pattern,
                    )
            else:
                yield from _traverse_dataframe_dict(
                    col["values"],
                    rules=rules,
                    curr_path=path_tail,
                    metrics=metrics,
                    prefix=next_prefix,
                )



from collections.abc import Generator

import msgspec
import polars as pl
from gcsfs import GCSFileSystem
from wcmatch import glob

from dapla_pseudo.v1.models.core import PseudoRule


class SchemaTraverser:
    """Perform transformations and operations on a potentially nested Parquet schema."""

    def __init__(self, schema: dict[str, pl.DataType], rules: list[PseudoRule]) -> None:
        """Initialize the class."""
        self.schema = schema
        self.rules = rules

    @staticmethod
    def from_path(schema_path: str, rules_path: str) -> "SchemaTraverser":
        """Build a SchemaTraverser instance."""
        schema_opener = (
            GCSFileSystem().open if schema_path.startswith("gs://") else open
        )
        rules_opener = GCSFileSystem().open if rules_path.startswith("gs://") else open
        with schema_opener(schema_path, mode="rb") as schema_file:
            schema = pl.read_parquet_schema(schema_file)
        with rules_opener(schema_path, mode="rb") as rules_file:
            content = msgspec.json.decode(rules_file.read())
            rules = [msgspec.convert(rule, PseudoRule) for rule in content]

        return SchemaTraverser(schema, rules)

    @staticmethod
    def write_rules(file_path: str, rules: list[PseudoRule]) -> None:
        """Write rules to file."""
        opener = GCSFileSystem.open if file_path.startswith("gs://") else open
        with opener(file_path, mode="wb") as rules_file:
            rules_file.write(msgspec.json.encode(rules))

    def match_rules(self, separator: str = "/") -> list[PseudoRule]:
        """Match a set of glob patterns.

        Takes a set of PseudoRules *without* concrete paths, and
        maps the glob-patterns to a concrete /path/to/thing using the schema.
        """
        return list(
            SchemaTraverser._match_rules(self.schema, self.rules, "", separator)
        )

    @staticmethod
    def _match_rules(
        schema: dict[str, pl.DataType],
        rules: list[PseudoRule],
        prev_path: str,
        separator: str,
    ) -> Generator[PseudoRule, None, None]:
        for name, dtype in schema.items():
            # Example for Polars types:
            # dtype     = List[Struct[Int64]]
            # base_type = List
            base_type = dtype.base_type()
            path = f"{prev_path}{separator}{name}".lstrip("/")  # remove leading /
            match base_type:
                case pl.Struct:  # If struct, nest further
                    yield from SchemaTraverser._match_rules(
                        dtype.to_schema(), rules, path, separator  # type: ignore[attr-defined]
                    )

                case pl.List | pl.Array:  # If list-like type, nest with the inner value
                    inner_type = dtype.inner  # type: ignore[attr-defined]
                    if inner_type == pl.Struct:
                        yield from SchemaTraverser._match_rules(
                            inner_type.to_schema(), rules, path, separator
                        )
                    elif inner_type == pl.List or inner_type == pl.Array:
                        raise ValueError(
                            f"Nested type: {dtype} could not be parsed.\
                            Nested types within nested types are not supported"
                        )
                    else:  # If not nested type, match on rule
                        if (
                            rule := SchemaTraverser._match_rule(path, rules)
                        ) is not None:
                            yield PseudoRule(
                                path=path,
                                pattern=rule.pattern,
                                func=rule.func,
                                name=rule.name,
                            )
                case _:
                    if (rule := SchemaTraverser._match_rule(path, rules)) is not None:
                        yield PseudoRule(
                            path=path,
                            pattern=rule.pattern,
                            func=rule.func,
                            name=rule.name,
                        )

    @staticmethod
    def _match_rule(path: str, rules: list[PseudoRule]) -> PseudoRule | None:
        for rule in rules:
            path = path.lstrip("/")  # remove leading /
            if glob.globmatch(
                path.lower(), rule.pattern.lower(), flags=glob.GLOBSTAR | glob.BRACE
            ):
                return rule
        return None



"""Builder for submitting a pseudonymization request."""

import typing as t
from datetime import date

import pandas as pd
import polars as pl
from dapla_metadata.datasets.core import Datadoc

from dapla_pseudo.constants import TIMEOUT_DEFAULT
from dapla_pseudo.constants import PredefinedKeys
from dapla_pseudo.constants import PseudoOperation
from dapla_pseudo.v1.baseclasses import _BasePseudonymizer
from dapla_pseudo.v1.baseclasses import _BaseRuleConstructor
from dapla_pseudo.v1.models.core import PseudoFunction
from dapla_pseudo.v1.models.core import PseudoKeyset
from dapla_pseudo.v1.models.core import PseudoRule
from dapla_pseudo.v1.result import Result


class Repseudonymize:
    """Starting point for pseudonymization of datasets.

    This class should not be instantiated, only the static methods should be used.
    """

    dataset: pl.DataFrame

    @staticmethod
    def from_pandas(dataframe: pd.DataFrame) -> "Repseudonymize._Repseudonymizer":
        """Initialize a pseudonymization request from a pandas DataFrame."""
        Repseudonymize.dataset = pl.from_pandas(dataframe)
        return Repseudonymize._Repseudonymizer()

    @staticmethod
    def from_polars(dataframe: pl.DataFrame) -> "Repseudonymize._Repseudonymizer":
        """Initialize a pseudonymization request from a polars DataFrame."""
        Repseudonymize.dataset = dataframe
        return Repseudonymize._Repseudonymizer()

    class _Repseudonymizer(_BasePseudonymizer):
        """Select one or multiple fields to be pseudonymized."""

        source_rules: t.ClassVar[list[PseudoRule]] = []
        target_rules: t.ClassVar[list[PseudoRule]] = []
        metadata: Datadoc | None = None

        def __init__(
            self,
            source_rules: list[PseudoRule] | None = None,
            target_rules: list[PseudoRule] | None = None,
            metadata: Datadoc | None = None,
        ) -> None:
            """Initialize the class."""
            if source_rules is None or target_rules is None:
                # Because the "source_rules" and "target_rules" are static class variables,
                # we need to reset the lists when the first call to "_Repseudonymizer" is made.
                # This is because if we were to use the base class "Repseudonymize" twice in the same file,
                # the lists of rules would persist across runs.
                # *This is very hacky*, but the alternative is to pass the rules through
                # "_RepseudoFuncSelectorSource" and "_RepseudoFuncSelectorTarget", which would hurt readability
                Repseudonymize._Repseudonymizer.source_rules = []
                Repseudonymize._Repseudonymizer.target_rules = []
            else:
                Repseudonymize._Repseudonymizer.source_rules.extend(source_rules)
                Repseudonymize._Repseudonymizer.target_rules.extend(target_rules)

            self.metadata = metadata

        def on_fields(
            self, *fields: str
        ) -> "Repseudonymize._RepseudoFuncSelectorSource":
            """Specify one or multiple fields to be pseudonymized."""
            return Repseudonymize._RepseudoFuncSelectorSource(
                list(fields), self.metadata
            )

        def with_metadata(self, metadata: Datadoc) -> "Repseudonymize._Repseudonymizer":
            """Specify existing datadoc metadata for the dataset."""
            return Repseudonymize._Repseudonymizer(metadata=metadata)

        def run(
            self,
            hierarchical: bool = False,
            source_custom_keyset: PseudoKeyset | str | None = None,
            target_custom_keyset: PseudoKeyset | str | None = None,
            timeout: int = TIMEOUT_DEFAULT,
        ) -> Result:
            """Pseudonymize the dataset.

            Args:
                hierarchical (bool): Whether the dataset is hierarchical or not. Needs PseudoRules with concrete paths. Defaults to False.
                source_custom_keyset (PseudoKeyset, optional): The source pseudonymization keyset to use. Defaults to None.
                target_custom_keyset (PseudoKeyset, optional): The target pseudonymization keyset to use. Defaults to None.
                timeout (int): The timeout in seconds for the API call. Defaults to TIMEOUT_DEFAULT.

            Returns:
                Result: The pseudonymized dataset and the associated metadata.
            """
            super().__init__(
                pseudo_operation=PseudoOperation.REPSEUDONYMIZE,
                dataset=Repseudonymize.dataset,
                hierarchical=hierarchical,
                user_provided_metadata=self.metadata,
            )

            result = super()._execute_pseudo_operation(
                rules=self.source_rules,
                target_rules=self.target_rules,
                custom_keyset=source_custom_keyset,
                target_custom_keyset=target_custom_keyset,
                timeout=timeout,
            )
            return result

    class _RepseudoFuncSelectorSource(_BaseRuleConstructor):
        def __init__(self, fields: list[str], metadata: Datadoc | None) -> None:
            self.fields = fields
            self._metadata = metadata
            super().__init__(fields)

        def from_stable_id(
            self,
            sid_snapshot_date: str | date | None = None,
            custom_key: PredefinedKeys | str | None = None,
        ) -> "Repseudonymize._RepseudoFuncSelectorTarget":
            """Claim that the selected fields were mapped to Stable ID, then pseudonymized with PAPIS-compatible encryption.

            Args:
                sid_snapshot_date (Optional[str | date], optional): Date representing SID-catalogue version that was used.
                    Latest if unspecified. Format: YYYY-MM-DD
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)

            Returns:
                An object with methods to choose how the field should be pseudonymized.
            """
            rules = super()._map_to_stable_id_and_pseudonymize(
                sid_snapshot_date, custom_key
            )
            return Repseudonymize._RepseudoFuncSelectorTarget(
                self.fields, rules, self._metadata
            )

        def from_default_encryption(
            self, custom_key: PredefinedKeys | str | None = None
        ) -> "Repseudonymize._RepseudoFuncSelectorTarget":
            """Claim that the selected fields were pseudonymized with default encryption.

            Args:
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)

            Returns:
                An object with methods to choose how the field should be pseudonymized.
            """
            rules = super()._with_daead_encryption(custom_key)
            return Repseudonymize._RepseudoFuncSelectorTarget(
                self.fields, rules, self._metadata
            )

        def from_papis_compatible_encryption(
            self, custom_key: PredefinedKeys | str | None = None
        ) -> "Repseudonymize._RepseudoFuncSelectorTarget":
            """Claim that the selected fields were pseudonymized with PAPIS-compatible encryption.

            Args:
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)

            Returns:
                An object with methods to choose how the field should be pseudonymized.
            """
            rules = super()._with_ff31_encryption(custom_key)
            return Repseudonymize._RepseudoFuncSelectorTarget(
                self.fields, rules, self._metadata
            )

        def from_custom_function(
            self, function: PseudoFunction
        ) -> "Repseudonymize._RepseudoFuncSelectorTarget":
            """Claim that the selected fields were pseudonymized with a custom, specified Pseudo Function."""
            rules = super()._with_custom_function(function)
            return Repseudonymize._RepseudoFuncSelectorTarget(
                self.fields, rules, self._metadata
            )

    class _RepseudoFuncSelectorTarget(_BaseRuleConstructor):
        def __init__(
            self,
            fields: list[str],
            source_rules: list[PseudoRule],
            metadata: Datadoc | None,
        ) -> None:
            self.source_rules = source_rules
            self._metadata = metadata
            super().__init__(fields)

        def to_stable_id(
            self,
            sid_snapshot_date: str | date | None = None,
            custom_key: PredefinedKeys | str | None = None,
        ) -> "Repseudonymize._Repseudonymizer":
            """Map the selected fields to Stable ID, then pseudonymize with a PAPIS-compatible encryption.

            In other words, this is a compound operation that both: 1) maps FNR to stable ID 2) then encrypts the Stable IDs.

            Args:
                sid_snapshot_date (Optional[str | date], optional): Date representing SID-catalogue version to use.
                    Latest if unspecified. Format: YYYY-MM-DD
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)

            Returns:
                Self: The object configured to be mapped to stable ID
            """
            rules = super()._map_to_stable_id_and_pseudonymize(
                sid_snapshot_date, custom_key
            )
            return Repseudonymize._Repseudonymizer(
                self.source_rules, rules, self._metadata
            )

        def to_default_encryption(
            self, custom_key: PredefinedKeys | str | None = None
        ) -> "Repseudonymize._Repseudonymizer":
            """Pseudonymize the selected fields with the default encryption algorithm (DAEAD).

            Args:
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (ssb-common-key-1)

            Returns:
                Self: The object configured to be mapped to stable ID
            """
            rules = super()._with_daead_encryption(custom_key)
            return Repseudonymize._Repseudonymizer(
                self.source_rules, rules, self._metadata
            )

        def to_papis_compatible_encryption(
            self, custom_key: PredefinedKeys | str | None = None
        ) -> "Repseudonymize._Repseudonymizer":
            """Pseudonymize the selected fields with a PAPIS-compatible encryption algorithm (FF31).

            Args:
                custom_key (Optional[PredefinedKeys | str], optional): Override the key to use for pseudonymization.
                    Must be one of the keys defined in PredefinedKeys. If not defined, uses the default key for this function (papis-common-key-1)

            Returns:
                Self: The object configured to be mapped to stable ID
            """
            rules = super()._with_ff31_encryption(custom_key)
            return Repseudonymize._Repseudonymizer(
                self.source_rules, rules, self._metadata
            )

        def to_custom_function(
            self, function: PseudoFunction
        ) -> "Repseudonymize._Repseudonymizer":
            rules = super()._with_custom_function(function)
            return Repseudonymize._Repseudonymizer(
                self.source_rules, rules, self._metadata
            )
